[CmdletBinding()]
param (
    [switch]
    $Defaults
)

$version = "V0.1"

Write-Host "Judas Vapoursynth Plugins Installer $version"
Write-Host "Checking for Python/Vapoursynth installation."

try {
    $pythonver = (python.exe --version).split(" ")[1].split(".")
    if ($pythonver[0] -eq 2) {
        Write-Host "Python 2 has been dead for years go away"
        exit
    }
    if ($pythonver[1] -lt 9) {
        Write-Host "Python must be at least version 3.9. Make sure that the correct one is in your path"
        exit
    }
    if ($pythonver[1] -gt 9) {
        Write-Host "WARNING: This script has been untested with Python 3.10+. It will probably still work, but no promises."
    }
    Write-Host "Found Python $($pythonver -join '.')"
}
catch {
    Write-Host "Couldn't find a Python install (Make sure python is in your path!)"
    exit
}

try {
    $vsver = (vspipe.exe -v)[2].split(" ")[1].substring(1)
    if ($vsver -lt 57) {
        Write-Host "Your Vapoursynth installation is too old (R$vsver), please upgrade to at least R57"
        exit
    }
    Write-Host "Found Vapoursynth R$vsver"
}
catch {
    Write-Host "Couldn't find a Vapoursynth install (Make sure vspipe is in your path!)"
    exit
}

python.exe -m pip >$null
if ($LASTEXITCODE) {
    Write-Host "Couldn't find a pip install corresponding to your python installation. Installing for you"
    (Invoke-WebRequest "https://bootstrap.pypa.io/get-pip.py").Content | python.exe
}

if (-not (Get-Command vsrepo.py -ErrorAction SilentlyContinue)) {
    Write-Host "Couldn't find vsrepo, make it is in your path"
}

$vsrepo = "python.exe `"$((Get-Command vsrepo.py).Source)`""
$sitepackages = (Invoke-Expression "$vsrepo paths" | Select-String 'Scripts: (.*)').Matches[0].Groups[1].Value
$plugins = (Invoke-Expression "$vsrepo paths" | Select-String 'Binaries: (.*)').Matches[0].Groups[1].Value

$pipinstall = "python.exe -m pip install --upgrade --user"

if (-not $Defaults) {
    $in = Read-Host "Using $sitepackages for scripts, specify different path, or none to continue"
    if (-not ([string]::IsNullOrWhiteSpace($in))) {
        if (-not (Test-Path $in -IsValid)) {
            Write-Host "$in isn't a valid path"
            exit
        }
        elseif (Test-Path $in -PathType Leaf) {
            Write-Host "$in is a file"
            exit
        }
        $sitepackages = $in
        $pipinstall += " --target $sitepackages"
    }

    $in = Read-Host "Using $plugins for plugins, specify different path, or none to continue"
    if (-not ([string]::IsNullOrWhiteSpace($in))) {
        if (-not (Test-Path $in -IsValid)) {
            Write-Host "$in isn't a valid path"
            exit
        }
        elseif (Test-Path $in -PathType Leaf) {
            Write-Host "$in is a file"
            exit
        }
        $plugins = $in
    }
}

Write-Host "Using $sitepackages for scripts"
Write-Host "Using $plugins for plugins"

New-Item -ItemType Directory -Force -Path $sitepackages | Out-Null
New-Item -ItemType Directory -Force -Path $plugins | Out-Null

Write-Host "Installing pip packages"
Invoke-Expression "$pipinstall -r requirements.txt"

Write-Host "Installing vsrepo packages"
Invoke-Expression "$vsrepo install (Get-Content .\vsrepo.txt) -b $plugins -s $sitepackages"

Write-Host "Copying extra plugins and scripts to relevant locations"
Copy-Item -Path .\scripts\* -Destination $sitepackages -Force
Copy-Item -Path .\plugins\* -Destination $plugins -Force

if (-not $Defaults) {
    $nvidia = (Get-CimInstance win32_VideoController).AdapterCompatibility -match "NVIDIA"
    if ($nvidia) {
        Write-Host "Nvidia GPU detected, auto installing CUDA plugins"
    }
    else {
        $in = Read-Host "No Nvidia GPU detected, do you still want CUDA plugins to be installed? (y/n)"
        $nvidia = [string]::Equals($in.ToLower(), "y")
    }
}

if ($nvidia) {
    Write-Host "Installing Nvidia vsrepo packages"
    Invoke-Expression "$vsrepo install (Get-Content .\vsrepo_nvidia.txt) -b $plugins -s $sitepackages"
    Write-Host "Copying extra Nvidia plugins and scripts to relevant locations"
    Copy-Item -Path .\plugins_nvidia\* -Destination $plugins -Force
}

Write-Host "All done!"